--- old/usr/src/uts/common/fs/zfs/dbuf.c	2014-10-28 11:57:18.883417588 -0700
+++ new/usr/src/uts/common/fs/zfs/dbuf.c	2014-10-28 11:57:18.886770577 -0700
@@ -1021,7 +1021,15 @@
 }
 
 dbuf_dirty_record_t *
-dbuf_dirty(dmu_buf_impl_t *db, dmu_tx_t *tx)
+dbuf_zero_dirty(dmu_buf_impl_t *db, dmu_tx_t *tx)
+{
+	ASSERT(db->db_objset != NULL);
+
+	return (dbuf_dirty(db, tx, B_TRUE));
+}
+
+dbuf_dirty_record_t *
+dbuf_dirty(dmu_buf_impl_t *db, dmu_tx_t *tx, boolean_t zero_write)
 {
 	dnode_t *dn;
 	objset_t *os;
@@ -1152,6 +1160,7 @@
 	 * transaction group won't leak out when we sync the older txg.
 	 */
 	dr = kmem_zalloc(sizeof (dbuf_dirty_record_t), KM_SLEEP);
+	dr->dr_zero_write = zero_write;
 	if (db->db_level == 0) {
 		void *data_old = db->db_buf;
 
@@ -1265,7 +1274,7 @@
 		if (drop_struct_lock)
 			rw_exit(&dn->dn_struct_rwlock);
 		ASSERT3U(db->db_level+1, ==, parent->db_level);
-		di = dbuf_dirty(parent, tx);
+		di = dbuf_dirty(parent, tx, B_FALSE);
 		if (parent_held)
 			dbuf_rele(parent, FTAG);
 
@@ -1408,7 +1417,7 @@
 		rf |= DB_RF_HAVESTRUCT;
 	DB_DNODE_EXIT(db);
 	(void) dbuf_read(db, NULL, rf);
-	(void) dbuf_dirty(db, tx);
+	(void) dbuf_dirty(db, tx, B_FALSE);
 }
 
 void
@@ -1435,7 +1444,25 @@
 	    dmu_tx_private_ok(tx));
 
 	dbuf_noread(db);
-	(void) dbuf_dirty(db, tx);
+	(void) dbuf_dirty(db, tx, B_FALSE);
+}
+
+
+void
+dmu_buf_will_zero_fill(dmu_buf_t *db_fake, dmu_tx_t *tx)
+{
+	dmu_buf_impl_t *db = (dmu_buf_impl_t *)db_fake;
+
+	ASSERT(db->db_blkid != DMU_BONUS_BLKID);
+	ASSERT(tx->tx_txg != 0);
+	ASSERT(db->db_level == 0);
+	ASSERT(!refcount_is_zero(&db->db_holds));
+
+	ASSERT(db->db.db_object != DMU_META_DNODE_OBJECT ||
+	    dmu_tx_private_ok(tx));
+
+	dbuf_noread(db);
+	(void) dbuf_zero_dirty(db, tx);
 }
 
 #pragma weak dmu_buf_fill_done = dbuf_fill_done
@@ -1520,7 +1547,7 @@
 	if (db->db_state == DB_CACHED &&
 	    refcount_count(&db->db_holds) - 1 > db->db_dirtycnt) {
 		mutex_exit(&db->db_mtx);
-		(void) dbuf_dirty(db, tx);
+		(void) dbuf_dirty(db, tx, B_FALSE);
 		bcopy(buf->b_data, db->db.db_data, db->db.db_size);
 		VERIFY(arc_buf_remove_ref(buf, db));
 		xuio_stat_wbuf_copied();
@@ -1551,7 +1578,7 @@
 	dbuf_set_data(db, buf);
 	db->db_state = DB_FILL;
 	mutex_exit(&db->db_mtx);
-	(void) dbuf_dirty(db, tx);
+	(void) dbuf_dirty(db, tx, B_FALSE);
 	dmu_buf_fill_done(&db->db, tx);
 }
 
@@ -2820,6 +2847,16 @@
 	dmu_write_policy(os, dn, db->db_level, wp_flag, &zp);
 	DB_DNODE_EXIT(db);
 
+	if (dr->dr_zero_write) {
+		zp.zp_zero_write = B_TRUE;
+
+		if (!spa_feature_is_active(os->os_spa, SPA_FEATURE_SPACE_RESERVATION))
+		{
+			spa_feature_incr(os->os_spa,
+			    SPA_FEATURE_SPACE_RESERVATION, tx);
+		}
+	}
+
 	if (db->db_level == 0 &&
 	    dr->dt.dl.dr_override_state == DR_OVERRIDDEN) {
 		/*
